/*
 * Extract files from a tar archive.
 *
 * Written 19 Nov 1985 by John Gilmore, ihnp4!hoptoad!gnu.
 *
 * @(#) extract.c 1.32 87/11/11 Public Domain - gnu
 */

#include <stdio.h>

#include "tar.h"

extern int errno;			/* From libc.a */
extern time_t time();			/* From libc.a */
extern char *index();			/* From libc.a or port.c */

extern union record *head;		/* Points to current tape header */
extern struct stat hstat;		/* Stat struct corresponding */
extern int head_standard;		/* Tape header is in ANSI format */

extern void print_header();
extern void skip_file();
extern void pr_mkdir();

int make_dirs();			/* Makes required directories */

static time_t now = 0;			/* Current time */
static we_are_root = 0;			/* True if our effective uid == 0 */
static int notumask = ~0;		/* Masks out bits user doesn't want */

#ifdef	VMS
char vmsname[NAMSIZ];			/* temp for making VMS filenames */
#endif

/*
 * Set up to extract files.
 */
extr_init()
{
	int ourmask;

	now = time((time_t *)0);
	if (geteuid() == 0)
		we_are_root = 1;

	/*
	 * We need to know our umask.  But if f_use_protection is set,
	 * leave our kernel umask at 0, and our "notumask" at ~0.
	 */
	ourmask = umask(0);		/* Read it */
	if (!f_use_protection) {
		(void) umask (ourmask);	/* Set it back how it was */
		notumask = ~ourmask;	/* Make umask override permissions */
	}
}


/*
 * Extract a file from the archive.
 */
void
extract_archive()
{
	register char *data;
	int fd, check, namelen, written, openflag, datain, dataout;
	long size, extno, exttotal, efsize;
	time_t acc_upd_times[2];
	char *myname, *mylinkname;
	char shortname[NAMSIZ+1], shortlinkname[NAMSIZ+1];

	saverec(&head);			/* Make sure it sticks around */
	userec(head);			/* And go past it in the archive */
	decode_header(head, &hstat, &head_standard, 1);	/* Snarf fields */
	getextents(&extno, &exttotal, &efsize);

	/* Print the record from 'head' and 'hstat' */
	if (f_verbose)
		print_header(stdout);

	/* truncate filenames to 14 characters */
	myname = head->header.name;
	if (f_shortnames) {
		if (truncname(myname, shortname))
			fprintf(stdout, "[ %s truncated to %s ]\n",
				myname, shortname);
	} else {
		strncpy(shortname, myname, NAMSIZ);
		shortname[NAMSIZ] = '\0';
	}
	myname = shortname;

	/*
	 * Check for fully specified pathnames and other atrocities.
	 *
	 * Note, we can't just make a pointer to the new file name,
	 * since saverec() might move the header and adjust "head".
	 * We have to start from "head" every time we want to touch
	 * the header record.
	 */
	if (f_absolute)
	    while ('/' == *myname) {
		static int warned_once = 0;

		myname++;			/* Force relative path */
		if (!warned_once++) {
			annorec(stderr, tar);
			fprintf(stderr,
	"Removing leading / from absolute path names in the archive.\n");
		}
	}

	/* show extent information */
	if (exttotal > 1L) {
		if (extno == 1L)
			fprintf(stdout,
			  "tar: large file %s needs %ld extents.\n",
			   myname, exttotal);
		fprintf(stdout, " +++ %s %ldK [extent #%ld of %ld]\n",
			myname, hstat.st_size / 1024L,
			extno, exttotal);
	}

	switch (head->header.linkflag) {

	default:
		annofile(stderr, tar);
		fprintf(stderr,
		   "Unknown file type '%c' for %s, extracted as normal file\n",
			head->header.linkflag, myname);
		/* FALL THRU */

	case LF_OLDNORMAL:
	case LF_NORMAL:
	case LF_CONTIG:
		/*
		 * Appears to be a file.
		 * See if it's really a directory.
		 */
		namelen = strlen(myname)-1;
		if (namelen > 0 && myname[namelen] == '/')
			goto really_dir;
#ifdef MSDOS
		if (namelen > 0 && myname[namelen] == '\\')
			goto really_dir;
#endif

		/* FIXME, deal with protection issues */
	again_file:
#ifdef VMS
		strncpy(vmsname, myname, NAMSIZ-1);
		vmsname[NAMSIZ-1] = '\0';
		vmsfixname(vmsname, 0);
		myname = vmsname;
#endif
		if (exttotal <= 1L) efsize = hstat.st_size;
		if (extno > 1L)
		  openflag = 
			O_BINARY|O_NDELAY|O_WRONLY|O_APPEND;
		else
		  openflag = f_keep?
			O_BINARY|O_NDELAY|O_WRONLY|O_APPEND|O_CREAT|O_EXCL:
			O_BINARY|O_NDELAY|O_WRONLY|O_APPEND|O_CREAT|O_TRUNC;
#ifdef VMS
		if (extno <= 1L) openflag &= ~O_APPEND;
		if ((f_cr & 4) == 4)
			fd = creat(myname, hstat.st_mode,
				"mrs=512", "rat=cr", "rfm=var");
		else
			fd = open(myname, openflag&~O_NDELAY,
				hstat.st_mode, "rfm=stmlf");
#else
#ifdef O_CTG
		/*
		 * Contiguous files (on the Masscomp) have to specify
		 * the size in the open call that creates them.
		 */
		if (head->header.lnkflag == LF_CONTIG)
			fd = open(myname, openflag | O_CTG,
				hstat.st_mode, efsize);
		else
#endif
		if (extno > 1L) {
			/* part of an extent. just paste it on... */
			fd = open(myname, openflag, hstat.st_mode);
		} else {
#ifdef NO_OPEN3
			/*
			 * On raw V7 we won't let them specify -k (f_keep), but
			 * we just bull ahead and create the files.
			 */
			fd = creat(myname, hstat.st_mode);
#else
			/*
			 * With 3-arg open(), we can do this up right.
			 */
			fd = open(myname, openflag, hstat.st_mode);
#endif
		}
#endif

		if (fd < 0) {
			if (make_dirs(myname))
				goto again_file;
			annofile(stderr, tar);
			fprintf(stderr, "Could not make file ");
			perror(myname);
			skip_file((long)hstat.st_size);
			goto quit;
		}

		for (size = hstat.st_size;
		     size > 0;
		     size -= written) {
			/*
			 * Locate data, determine max length
			 * writeable, write it, record that
			 * we have used the data, then check
			 * if the write worked.
			 */
			data = findrec()->charptr;
			if (data == NULL) {	/* Check it... */
				annorec(stderr, tar);
				fprintf(stderr, "Unexpected EOF on archive file\n");
				break;
			}
			written = endofrecs()->charptr - data;
			if (written > size) written = size;
			errno = 0;
			if ((f_cr & 1) == 1) {
				datain = dataout = 0;
				while (datain < written) {
					if (data[datain] == '\015') ++datain;
					else ar_crblock[dataout++] =
					     data[datain++];
				}
				check = write(fd, ar_crblock, dataout);
				if (check == dataout) check = written;
			} else if ((f_cr & 2) == 2) {
				datain = dataout = 0;
				while (datain < written) {
					if (data[datain] == '\n')
					  ar_crblock[dataout++] = '\015';
					ar_crblock[dataout++] =
					     data[datain++];
				}
				check = write(fd, ar_crblock, dataout);
				if (check == dataout) check = written;
			}
			else check = write(fd, data, written);
			/*
			 * The following is in violation of strict
			 * typing, since the arg to userec
			 * should be a struct rec *.  FIXME.
			 */
			userec((union record *)(data + written - 1));
			if (check == written) continue;
			/*
			 * Error in writing to file.
			 * Print it, skip to next file in archive.
			 */
			annofile(stderr, tar);
			fprintf(stderr,
	"Tried to write %d bytes to file, could only write %d:\n",
				written, check);
			perror(myname);
			skip_file((long)(size - written));
			break;	/* Still do the close, mod time, chmod, etc */
		}

		check = close(fd);
		if (check < 0) {
			annofile(stderr, tar);
			fprintf(stderr, "Error while closing ");
			perror(myname);
		}
		
	set_filestat:
		/*
		 * Set the modified time of the file.
		 * 
		 * Note that we set the accessed time to "now", which
		 * is really "the time we started extracting files".
		 */
		if (!f_modified) {
			acc_upd_times[0] = now;	         /* Accessed now */
			acc_upd_times[1] = hstat.st_mtime; /* Mod'd */
			if (utime(myname,
			    acc_upd_times) < 0) {
				annofile(stderr, tar);
				fprintf(stderr, "Error while updating time ");
				perror(myname);
			}
		}

		/*
		 * If we are root, set the owner and group of the extracted
		 * file.  This does what is wanted both on real Unix and on
		 * System V.  If we are running as a user, we extract as that
		 * user; if running as root, we extract as the original owner.
		 */
		if (we_are_root) {
			if (chown(myname, hstat.st_uid,
				  hstat.st_gid) < 0) {
				annofile(stderr, tar);
				fprintf(stderr, "Error while updating owner ");
				perror(myname);
			}
		}

		/*
		 * If '-k' is not set, open() or creat() could have saved
		 * the permission bits from a previously created file,
		 * ignoring the ones we specified.
		 * Even if -k is set, if the file has abnormal
		 * mode bits, we must chmod since writing or chown() has
		 * probably reset them.
		 *
		 * If -k is set, we know *we* created this file, so the mode
		 * bits were set by our open().   If the file is "normal", we
		 * skip the chmod.  This works because we did umask(0) if -p
		 * is set, so umask will have left the specified mode alone.
		 */
		if ((!f_keep)
		    || (hstat.st_mode & (S_ISUID|S_ISGID|S_ISVTX))) {
			if (chmod(myname,
				  notumask & (int)hstat.st_mode) < 0) {
				annofile(stderr, tar);
				fprintf(stderr, "Error while updating mode ");
				perror(myname);
			}
		}

	quit:
		break;

	case LF_LINK:
	again_link:
		mylinkname = head->header.linkname;
		if (f_shortnames) {
			if (truncname(mylinkname, shortlinkname))
				fprintf(stdout, "[ link %s truncated to %s ]\n",
					mylinkname, shortlinkname);
			mylinkname = shortlinkname;
		}
		check = link (mylinkname, myname);
		if (check == 0)
			break;
		if (make_dirs(myname))
			goto again_link;
		annofile(stderr, tar);
		fprintf(stderr, "Could not link %s to ",
			myname);
		perror(head->header.linkname);
		break;

#ifdef S_IFLNK
	case LF_SYMLINK:
	again_symlink:
		mylinkname = head->header.linkname;
		if (f_shortnames) {
			if (truncname(mylinkname, shortlinkname))
				fprintf(stdout, "[ link %s truncated to %s ]\n",
					mylinkname, shortlinkname);
			mylinkname = shortlinkname;
		}
		check = symlink(mylinkname, myname);
		/* FIXME, don't worry uid, gid, etc... */
		if (check == 0)
			break;
		if (make_dirs(myname))
			goto again_symlink;
		annofile(stderr, tar);
		fprintf(stderr, "Could not create symlink ");
		perror(head->header.linkname);
		break;
#endif

#ifdef S_IFCHR
	case LF_CHR:
		hstat.st_mode |= S_IFCHR;
		goto make_node;
#endif

#ifdef S_IFBLK
	case LF_BLK:
		hstat.st_mode |= S_IFBLK;
		goto make_node;
#endif

#ifdef S_IFIFO
	/* If local system doesn't support FIFOs, use default case */
	case LF_FIFO:
		hstat.st_mode |= S_IFIFO;
		hstat.st_rdev = 0;		/* FIXME, do we need this? */
		goto make_node;
#endif

	make_node:
		check = mknod(myname,
			      (int) hstat.st_mode, (int) hstat.st_rdev);
		if (check != 0) {
			if (make_dirs(myname))
				goto make_node;
			annofile(stderr, tar);
			fprintf(stderr, "Could not make ");
			perror(myname);
			break;
		};
		goto set_filestat;

	case LF_DIR:
		namelen = strlen(myname)-1;
	really_dir:
		/* Check for trailing /, and zap as many as we find. */
		while (namelen && myname[namelen] == '/')
			myname[namelen--] = '\0';
#ifdef MSDOS
		while (namelen && myname[namelen] == '\\')
			myname[namelen--] = '\0';
#endif
		
	again_dir:
#ifdef VMS
		strncpy(vmsname, myname, NAMSIZ-1);
		vmsname[strlen(vmsname)] = '/';
		vmsname[NAMSIZ-1] = '\0';
		vmsfixname(vmsname, 1);
		check = vmsmkdir(vmsname, 0300 | (int)hstat.st_mode);
#else
		check = mkdir(myname,
			      0300 | (int)hstat.st_mode);
#endif
		if (check != 0) {
			if (make_dirs(myname))
				goto again_dir;
			/* If we're trying to create '.', let it be. */
			if (myname[namelen] == '.' && 
			    (namelen==0 ||
			     myname[namelen-1]=='/'))
				goto check_perms;
			annofile(stderr, tar);
			fprintf(stderr, "Could not make directory ");
			perror(myname);
			break;
		}
		
	check_perms:
		if (0300 != (0300 & (int) hstat.st_mode)) {
			hstat.st_mode |= 0300;
			annofile(stderr, tar);
			fprintf(stderr,
			  "Added write & execute permission to directory %s\n",
			  myname);
		}

		goto set_filestat;
		/* FIXME, Remember timestamps for after files created? */
		/* FIXME, change mode after files created (if was R/O dir) */

	}

	/* We don't need to save it any longer. */
	saverec((union record **) 0);	/* Unsave it */
}

/*
 * After a file/link/symlink/dir creation has failed, see if
 * it's because some required directory was not present, and if
 * so, create all required dirs.
 */
int
make_dirs(pathname)
	char *pathname;
{
	char *p;			/* Points into path */
	int madeone = 0;		/* Did we do anything yet? */
	int save_errno = errno;		/* Remember caller's errno */
	int check;

	if (errno != ENOENT)
		return 0;		/* Not our problem */

	for (p = index(pathname, '/'); p != NULL; p = index(p+1, '/')) {
		/* Avoid mkdir of empty string, if leading or double '/' */
		if (p == pathname || p[-1] == '/')
			continue;
		/* Avoid mkdir where last part of path is '.' */
		if (p[-1] == '.' && (p == pathname+1 || p[-2] == '/'))
			continue;
		*p = 0;				/* Truncate the path there */
#ifdef VMS
		strncpy(vmsname, pathname, NAMSIZ-1);
		vmsname[NAMSIZ-1] = '\0';
		vmsfixname(vmsname, 1);
		check = vmsmkdir(vmsname, 0777);
#else
		check = mkdir (pathname, 0777);	/* Try to create it as a dir */
#endif
		if (check == 0) {
			/* Fix ownership */
			if (we_are_root) {
#ifdef VMS
				strncpy(vmsname, pathname, NAMSIZ-1);
				vmsname[NAMSIZ-1] = '\0';
				vmsfixname(vmsname, 0);
				vmsname[NAMSIZ-5] = '\0';
				strcat(vmsname, ".dir");
				if (chown(vmsname,
#else
				if (chown(pathname,
#endif
					hstat.st_uid, hstat.st_gid) < 0) {
					annofile(stderr, tar);
					fprintf(stderr,
						"Error while updating owner ");
					perror(pathname);
				}
			}
			pr_mkdir(pathname, p-pathname, notumask&0777, stdout);
			madeone++;		/* Remember if we made one */
			*p = '/';
			continue;
		}
		*p = '/';
		if (errno == EEXIST)		/* Directory already exists */
			continue;
		/*
		 * Some other error in the mkdir.  We return to the caller.
		 */
		break;
	}

	errno = save_errno;		/* Restore caller's errno */
	return madeone;			/* Tell them to retry if we made one */
}
